Package: qgam
Type: Package
Title: Smooth Additive Quantile Regression Models
Version: 2.0.0
Date: 2025-04-10
Authors@R: c(person("Matteo", "Fasiolo", email = "matteo.fasiolo@gmail.com", role = c("aut", "cre")),
             person("Ben", "Griffiths", role = c("aut")), 
             person("Simon", "N. Wood", role = c("ctb")), 
             person("Margaux", "Zaffran", role = c("ctb")), 
             person("Yannig", "Goude", role = c("ctb")),
             person("Raphael", "Nedellec", role = c("ctb")))
Maintainer: Matteo Fasiolo <matteo.fasiolo@gmail.com>
Description: Smooth additive quantile regression models, fitted using
    the methods of Fasiolo et al. (2020) <doi:10.1080/01621459.2020.1725521>.
    See Fasiolo at al. (2021) <doi:10.18637/jss.v100.i09> for an introduction to the package. Differently from
    'quantreg', the smoothing parameters are estimated automatically by marginal
    loss minimization, while the regression coefficients are estimated using either
    PIRLS or Newton algorithm. The learning rate is determined so that the Bayesian
    credible intervals of the estimated effects have approximately the correct
    coverage. The main function is qgam() which is similar to gam() in 'mgcv', but
    fits non-parametric quantile regression models.
License: GPL (>= 2)
Depends: R (>= 4.0), mgcv (>= 1.9)
Imports: shiny, plyr, doParallel, parallel, grDevices
Suggests: knitr, rmarkdown, MASS, RhpcBLASctl, testthat
VignetteBuilder: knitr
RoxygenNote: 7.3.2
Encoding: UTF-8
NeedsCompilation: yes
Packaged: 2025-04-10 13:23:38 UTC; mf15002
Author: Matteo Fasiolo [aut, cre],
  Ben Griffiths [aut],
  Simon N. Wood [ctb],
  Margaux Zaffran [ctb],
  Yannig Goude [ctb],
  Raphael Nedellec [ctb]
Repository: CRAN
Date/Publication: 2025-04-10 19:20:10 UTC
Built: R 4.3.3; x86_64-apple-darwin20; 2025-04-10 22:11:27 UTC; unix
Archs: qgam.so.dSYM
